from qq.plugins.sqlalchemy.exceptions import SettingMissing


class TestSettingMissing:
    def test_stringify(self):
        """
        SettingMissing should be able to format proper message.
        """
        assert str(SettingMissing('key', 'name')) == "'key' key is needed for use 'name' database in application"



from unittest.mock import MagicMock

from pytest import fixture
from pytest import raises

from qq import ApplicationNotStartedError
from qq import Context
from qq import InitializeInjectors
from qq import SimpleInjector
from qq.application import Application
from qq.injector import Injector
from qq.injector import get_injectors_from_function
from qq.injector import is_injector_ready
from qq.plugins import SettingsPlugin


class TestIsInjectorReady:
    @fixture
    def mparameter(self):
        return MagicMock()

    @fixture
    def mbound_args(self):
        args = MagicMock()
        args.arguments = {}
        return args

    def test_when_parameter_is_not_an_injector(self, mparameter, mbound_args):
        assert is_injector_ready(mparameter, "key", mbound_args) is False

    def test_when_parameter_is_already_provided(self, mparameter, mbound_args):
        mbound_args.arguments["key"] = 1
        mparameter._default = Injector(None)

        assert is_injector_ready(mparameter, "key", mbound_args) is False

    def test_when_parameter_is_not_provided(self, mparameter, mbound_args):
        mparameter._default = Injector(None)
        result = is_injector_ready(mparameter, "key", mbound_args)
        assert result is True


app = Application()


def example_fun(first, second, third=SimpleInjector(app, "key")):
    return [first, second, third]


class TestInjectors:
    def test_when_arguments_provided(self):
        assert list(get_injectors_from_function(example_fun, [1, 2, 3], {})) == []

    def test_when_arguments_not_provided(self):
        """
        When injector argument not provided, injectors fun should return this
        injector.
        """
        context = Context(Application())
        context.values["key"] = 333

        name, value = list(get_injectors_from_function(example_fun, [1, 2], {}))[0]

        assert name == "third"
        assert value(context, "key") != value


def default_settings():
    return {"settings": True}


class ExampleApplication(Application):
    def create_plugins(self):
        self.plugins["settings"] = SettingsPlugin("qq.tests.test_injector")


class TestInitializeInjectors:
    @fixture
    def app(self):
        return ExampleApplication()

    @fixture
    def fun(self, app):
        def example_fun(first, second, third=SimpleInjector("settings")):
            return [first, second, third]

        return InitializeInjectors(app)(example_fun)

    def test_when_arguments_provided(self, fun):
        assert fun(1, 2, 3) == [1, 2, 3]

    def test_when_arguments_not_provided_and_app_not_started(self, fun):
        with raises(ApplicationNotStartedError):
            fun(1, 2)

    def test_when_arguments_not_provided_and_app_started(self, app, fun):
        app.start("default_settings")
        assert fun(1, 2) == [1, 2, {"settings": True}]



import json
from dataclasses import dataclass
from datetime import date
from datetime import datetime
from decimal import Decimal
from enum import Enum
from unittest.mock import MagicMock
from uuid import UUID
from uuid import uuid4

from pytest import fixture
from pytest import raises

from qq.plugins.jsonhack.encodergenerators import encoder_for
from qq.plugins.jsonhack.encoders import DateEncoder
from qq.plugins.jsonhack.encoders import DatetimeEncoder
from qq.plugins.jsonhack.encoders import DecimalEncoder
from qq.plugins.jsonhack.encoders import UUIDEncoder
from qq.plugins.jsonhack.errors import UnknownObjectError
from qq.plugins.jsonhack.models import ENCODERS_CACHE
from qq.plugins.jsonhack.models import add_encoder
from qq.plugins.jsonhack.models import init_encoders
from qq.plugins.jsonhack.models import object_hook
from qq.plugins.jsonhack.stubs import DecoderStub
from qq.plugins.jsonhack.stubs import EncoderStub
from qq.plugins.types import CustomBaseType


class CustomUuid(UUID, CustomBaseType):
    pass


class SampleEnum(Enum):
    FIRST = "first"
    SECOND = "second"


@dataclass
class SampleDataclass:
    name: str
    year: int


data = {
    "0": 0,
    "1": datetime.now(),
    "2": uuid4(),
    "3": date.today(),
    "4": Decimal("3.14"),
    "5": SampleDataclass(name="first", year=2019),
    "6": CustomUuid(uuid4().hex),
    "7": SampleEnum("first"),
    "nested": {
        "n0": 0,
        "n1": datetime.now(),
        "n2": uuid4(),
        "n3": date.today(),
        "n4": Decimal("6.14"),
        "n5": SampleDataclass(name="second", year=2020),
        "n6": CustomUuid(uuid4().hex),
        "n7": SampleEnum("second"),
    },
}


class FakeClass:
    pass


class TestEncodingJson:
    @fixture(autouse=True)
    def stub(self):
        encoder_stub = EncoderStub(json.JSONEncoder, "default")
        decoder_stub = DecoderStub(json.JSONDecoder, "decode")
        add_encoder(DatetimeEncoder())
        add_encoder(DateEncoder())
        add_encoder(UUIDEncoder())
        add_encoder(DecimalEncoder())
        for encoder in encoder_for([SampleDataclass, CustomUuid, SampleEnum]):
            add_encoder(encoder)

        encoder_stub.stub()
        decoder_stub.stub()
        yield
        encoder_stub.unstub()
        decoder_stub.unstub()
        ENCODERS_CACHE.clear()

    def test_normal(self):
        result = json.loads(json.dumps(data))

        assert data == result

    def test_failed_encoding(self):
        newdata = FakeClass()

        with raises(TypeError):
            json.loads(json.dumps(newdata))


class TestObjectHook:
    @fixture
    def mget_encoders(self, mocker, mencoder):
        mock = mocker.patch("qq.plugins.jsonhack.models.get_encoders")
        mock.return_value = [mencoder]
        return mock

    @fixture
    def mencoder(self):
        return MagicMock()

    @fixture
    def data(self):
        return {"_type": "something", "value": "here"}

    def test_when_normal_type(self, data):
        """
        object_hook should do nothing if the object to decode has no "_type" in
        keys
        """
        del data["_type"]
        assert object_hook(data) == data

    def test_when_encoded(self, mencoder, data, mget_encoders):
        """
        object_hook should decode the object if the encoder is found
        """
        assert object_hook(data) == mencoder.decode.return_value
        mencoder.decode.assert_called_once_with(data)
        mencoder.is_decodable.assert_called_once_with(data)

    def test_when_unknow_type(self, data, mget_encoders):
        """
        object_hook should raise an error if the encoder for the object is not found
        """
        mget_encoders.return_value = []
        with raises(UnknownObjectError):
            object_hook(data)


class TestInitEncoders:
    @fixture
    def finder(self):
        return MagicMock()

    @fixture(autouse=True)
    def clear_encoders(self):
        yield
        ENCODERS_CACHE.clear()

    def test_parse_encoders(self):
        init_encoders([])

        assert len(ENCODERS_CACHE) is not 0

    def test_with_finders(self, finder):
        init_encoders([finder])

        finder.find.return_value = [CustomUuid]

        init_encoders([finder])

        # I know this is strange, but I did not know how to test this behaviour
        # better
        for enc in ENCODERS_CACHE:
            if enc.TYPE == CustomUuid:
                assert True
                return

        assert False



from typing import Dict

from sqlalchemy.engine import Engine
from sqlalchemy.engine import create_engine
from sqlalchemy.engine.url import make_url
from sqlalchemy.orm import Session
from sqlalchemy.orm import sessionmaker

from qq.application import Application
from qq.context import Context
from qq.plugins.settings import SettingsBasedPlugin
from qq.plugins.sqlalchemy.consts import ENGINE_KEY
from qq.plugins.sqlalchemy.consts import OPTIONS_KEY
from qq.plugins.sqlalchemy.consts import SESSIONMAKER_KEY
from qq.plugins.sqlalchemy.consts import URL_KEY
from qq.plugins.sqlalchemy.exceptions import SettingMissing


class SqlAlchemyPlugin(SettingsBasedPlugin):
    @property
    def url(self):
        """
        Get url from settings.
        """
        return self._settings[URL_KEY]

    @property
    def dbname(self):
        return make_url(self.url).database

    def start(self, application: Application) -> Dict:
        self._settings = self.get_my_settings(application)
        self._validate_settings()
        self.engine = self.create_engine()
        self.sessionmaker = sessionmaker(
            autoflush=False, autocommit=False, bind=self.engine
        )
        return {
            ENGINE_KEY: self.engine,
            SESSIONMAKER_KEY: self.sessionmaker,
        }

    def enter(self, context: Context) -> Session:
        self.session = self.sessionmaker()
        return self.session

    def exit(self, context, exc_type, exc_value, traceback):
        if exc_type or self._settings.get("tests", False):
            self.session.rollback()
        self.session.close()

    def create_engine(self) -> Engine:
        return create_engine(self.url, **self._settings.get(OPTIONS_KEY, {}))

    def recreate(self, metadata):
        engine = self.create_engine()
        metadata.drop_all(engine)
        metadata.create_all(engine)

    def _validate_settings(self):
        if URL_KEY not in self._settings:
            raise SettingMissing(URL_KEY, self.key)
        make_url(self._settings[URL_KEY])



from abc import ABC
from abc import abstractmethod
from datetime import date
from datetime import datetime
from decimal import Decimal
from uuid import UUID

from dateutil import parser
from marshmallow_dataclass import class_schema


class Encoder(ABC):
    @property
    @abstractmethod
    def TYPE(self):
        pass  # pragma: no cover

    @property
    def TYPENAME(self):
        return (
            getattr(self.TYPE, "ENCODED_TYPENAME", None) or self.TYPE.__name__
        )

    def is_encodable(self, value):
        return type(value) == self.TYPE

    def is_decodable(self, obj):
        return obj.get("_type") == self.TYPENAME

    def encode(self, value):
        return {"_type": self.TYPENAME, "value": self._encode(value)}

    def decode(self, obj):
        return self._decode(obj["value"])

    @abstractmethod
    def _encode(self, value):
        pass  # pragma: no cover

    @abstractmethod
    def _decode(self, value):
        pass  # pragma: no cover


class DatetimeEncoder(Encoder):
    TYPENAME = "datetime"
    TYPE = datetime

    def _encode(self, value):
        return value.isoformat()

    def _decode(self, value):
        return parser.parse(value)


class UUIDEncoder(Encoder):
    TYPENAME = "UUID"
    TYPE = UUID

    def _encode(self, value):
        return value.hex

    def _decode(self, value):
        return UUID(value)


class DateEncoder(Encoder):
    TYPENAME = "date"
    TYPE = date

    def _encode(self, value):
        return value.isoformat()

    def _decode(self, value):
        return parser.parse(value).date()


class DecimalEncoder(Encoder):
    TYPENAME = "decimal"
    TYPE = Decimal

    def _encode(self, value):
        return str(value)

    def _decode(self, value):
        return Decimal(value)


class EnumEncoder(Encoder):
    def _encode(self, value):
        return value.value

    def _decode(self, value):
        return self.TYPE(value)


class DataclassEncoder(Encoder):
    def _get_schema(self):
        return class_schema(self.TYPE)()

    def _encode(self, obj):
        schema = self._get_schema()
        return schema.dump(obj)

    def _decode(self, data):
        schema = self._get_schema()
        return schema.load(data)



from pyramid.httpexceptions import HTTPMethodNotAllowed
from pyramid.request import Request
from pyramid.traversal import DefaultRootFactory


class View:
    def __init__(self, root_factory: DefaultRootFactory, request: Request):
        self.request = request
        self.root_factory = root_factory

    @property
    def methods(self):
        return {
            "GET": self.get,
            "POST": self.post,
            "PUT": self.put,
            "PATCH": self.patch,
            "DELETE": self.delete,
            "OPTIONS": self.options,
        }

    def __call__(self):
        method = self.methods[self.request.method]
        return method(self.request)

    def get(self, request: Request):
        raise HTTPMethodNotAllowed()

    def post(self, request: Request):
        raise HTTPMethodNotAllowed()

    def put(self, request: Request):
        raise HTTPMethodNotAllowed()

    def patch(self, request: Request):
        raise HTTPMethodNotAllowed()

    def delete(self, request: Request):
        raise HTTPMethodNotAllowed()

    def options(self, request: Request):
        raise HTTPMethodNotAllowed()


class RestfulView(View):
    """
    View which will return context as json.
    """

    renderer = "json"



from unittest.mock import MagicMock
from unittest.mock import sentinel

from pytest import fixture

from qq.plugins.redis import SETTINGS_DB_KEY
from qq.plugins.redis import SETTINGS_HOST_KEY
from qq.plugins.redis import SETTINGS_PORT_KEY
from qq.plugins.redis import RedisPlugin

PREFIX = "qq.plugins.redis"


class TestRedisPlugin:
    @fixture
    def plugin(self):
        plugin = RedisPlugin()
        plugin.init("redis")
        return plugin

    @fixture
    def context(self):
        return MagicMock()

    @fixture
    def mget_my_settings(self, mocker, plugin):
        mock = mocker.patch.object(plugin, "get_my_settings")
        mock.return_value = {
            SETTINGS_HOST_KEY: sentinel.host,
            SETTINGS_PORT_KEY: sentinel.port,
            SETTINGS_DB_KEY: sentinel.dbkey,
        }
        return mock

    @fixture
    def mredis(self, mocker):
        return mocker.patch(f"{PREFIX}.Redis")

    def test_connection(self, plugin, mredis, context, mget_my_settings):
        """
        Plugin should start redis connection when creating context.
        """
        assert plugin.enter(context) == mredis.return_value

        mredis.assert_called_once_with(
            host=sentinel.host, port=sentinel.port, db=sentinel.dbkey
        )



from unittest.mock import MagicMock
from unittest.mock import sentinel

from pytest import fixture

from qq.plugins.pyramid.plugins import AuthPlugin
from qq.plugins.pyramid.plugins import BasePyramidPlugin
from qq.plugins.pyramid.plugins import CsrfPlugin
from qq.plugins.pyramid.plugins import RoutingPlugin
from qq.plugins.pyramid.plugins import SessionPlugin


class Fixtures:
    @fixture
    def mpyramid(self):
        return MagicMock()


class TestBasePyramidPlugin(Fixtures):
    @fixture
    def mapp(self):
        app = MagicMock()
        app.globals = {
            "settings": {
                "dbkey": sentinel.dbsettings,
            },
        }
        return app

    @fixture
    def plugin(self):
        plugin = BasePyramidPlugin()
        plugin.init("dbkey")
        return plugin

    def test_getting_settings(self, mapp, plugin):
        plugin.start(mapp)

        assert plugin.settings == sentinel.dbsettings


class TestAuthPlugin(Fixtures):
    @fixture
    def mauthn_policy_cls(self):
        return MagicMock()

    @fixture
    def mauthz_policy_cls(self):
        return MagicMock()

    @fixture
    def mroot_factory(self):
        return MagicMock()

    @fixture
    def plugin(self, mauthn_policy_cls, mauthz_policy_cls, mroot_factory):
        return AuthPlugin(mauthn_policy_cls, mauthz_policy_cls, mroot_factory)

    def test_start_pyramid(
        self,
        plugin,
        mpyramid,
        mauthn_policy_cls,
        mauthz_policy_cls,
        mroot_factory,
    ):
        """
        .start_pyramid should configure auth policy
        """
        plugin.settings = {"secret": sentinel.secret}
        plugin.start_pyramid(mpyramid)

        mauthn_policy_cls.assert_called_once_with(sentinel.secret)
        mauthz_policy_cls.assert_called_once_with()

        mpyramid.set_authentication_policy.assert_called_once_with(
            mauthn_policy_cls.return_value
        )
        mpyramid.set_authorization_policy.assert_called_once_with(
            mauthz_policy_cls.return_value
        )
        mpyramid.set_root_factory.assert_called_once_with(mroot_factory)

    def test_start_pyramid_without_root_factory(
        self,
        plugin,
        mpyramid,
        mauthn_policy_cls,
        mauthz_policy_cls,
        mroot_factory,
    ):
        """
        .start_pyramid without root_factory should configure auth policy without
        root_factory
        """
        plugin.settings = {"secret": sentinel.secret}
        plugin.root_factory = None
        plugin.start_pyramid(mpyramid)

        mauthn_policy_cls.assert_called_once_with(sentinel.secret)
        mauthz_policy_cls.assert_called_once_with()

        mpyramid.set_authentication_policy.assert_called_once_with(
            mauthn_policy_cls.return_value
        )
        mpyramid.set_authorization_policy.assert_called_once_with(
            mauthz_policy_cls.return_value
        )
        assert not mpyramid.set_root_factory.called


class TestCsrfPlugin(Fixtures):
    @fixture
    def mpolicy_cls(self):
        return MagicMock()

    @fixture
    def plugin(self, mpolicy_cls):
        return CsrfPlugin(mpolicy_cls)

    def test_start_pyramid(self, plugin, mpyramid, mpolicy_cls):
        """
        .start_pyramid should configure csrf
        """
        plugin.settings = {
            "csrf_token_key": sentinel.csrf_token_key,
            "csrf_header_key": sentinel.csrf_header_key,
        }

        plugin.start_pyramid(mpyramid)

        mpolicy_cls.assert_called_once_with()
        mpyramid.set_csrf_storage_policy(mpolicy_cls.return_value)
        mpyramid.set_default_csrf_options.assert_called_once_with(
            require_csrf=True,
            token=sentinel.csrf_token_key,
            header=sentinel.csrf_header_key,
        )


class TestRoutingPlugin(Fixtures):
    @fixture
    def mrouting_cls(self):
        return MagicMock()

    @fixture
    def plugin(self, mrouting_cls):
        return RoutingPlugin(mrouting_cls)

    def test_start_pyramid(self, plugin, mpyramid, mrouting_cls):
        """
        .start_pyramid should start routing.
        """
        plugin.start_pyramid(mpyramid)
        mrouting_cls.assert_called_once_with(mpyramid)
        mrouting_cls.return_value.make.assert_called_once_with()


class TestSessionPlugin(Fixtures):
    @fixture
    def msession_factory_cls(self):
        return MagicMock()

    @fixture
    def plugin(self, msession_factory_cls):
        return SessionPlugin(msession_factory_cls)

    def test_start_pyramid(self, plugin, mpyramid, msession_factory_cls):
        """
        .start_pyramid should configure csrf
        """
        plugin.settings = {
            "session_secret": sentinel.session_secret,
        }

        plugin.start_pyramid(mpyramid)

        msession_factory_cls.assert_called_once_with(sentinel.session_secret)
        mpyramid.set_session_factory.assert_called_once_with(
            msession_factory_cls.return_value
        )



from unittest.mock import MagicMock
from unittest.mock import patch
from unittest.mock import sentinel

from pytest import fixture

from qq.plugins.pyramid.application import PyramidApplication
from qq.plugins.settings import SettingsPlugin

PREFIX = "qq.plugins.pyramid.application"


class ExampleApplication(PyramidApplication):
    def create_plugins(self):
        super().create_plugins()
        self.plugins["plugin1"] = MagicMock()
        self.plugins["plugin2"] = MagicMock()
        del self.plugins["plugin2"].start_pyramid


class TestPyramidApplication:
    @fixture
    def configurator(self):
        return ExampleApplication()

    @fixture
    def mpyramid_configurator(self):
        with patch(f"{PREFIX}.Configurator") as mock:
            yield mock

    def test_starting_pyramid_application(
        self, configurator, mpyramid_configurator
    ):
        """
        .start_pyramid should create wsgi application
        """
        configurator.start("pyramid")

        configurator.globals[SettingsPlugin.DEFAULT_KEY] = sentinel.settings
        wsgi = configurator.make_wsgi_app("arg", kw="arg2")

        mpyramid_configurator.assert_called_once_with(
            "arg", settings=sentinel.settings, kw="arg2"
        )
        pyramid = mpyramid_configurator.return_value
        pyramid.make_wsgi_app.assert_called_once_with()
        assert pyramid.make_wsgi_app.return_value == wsgi

        configurator.plugins["plugin1"].start_pyramid.assert_called_once_with(
            pyramid
        )



from sqlalchemy.orm import Session

from qq.context import Context
from qq.injector import ContextManagerInjector
from qq.injector import Injector
from qq.plugins.settings import TESTS_KEY
from qq.plugins.settings import SettingsPlugin


@Injector
def SAQuery(context: Context, key: str) -> Session:
    return context[key]


class SACommand(ContextManagerInjector):
    def __init__(
        self, key: str, settings_key: str = SettingsPlugin.DEFAULT_KEY
    ):
        super().__init__()
        self.key = key
        self.settings_key = settings_key

    def __enter__(self, context) -> Session:
        return context[self.key]

    def __exit__(
        self,
        exc_type,
        exc_value,
        traceback,
        context,
    ):
        settings = context[self.settings_key][self.key]
        db = context[self.key]
        if settings.get(TESTS_KEY, False):
            db.flush()
        elif not exc_type:
            db.commit()



from celery import Celery

from qq.application import Application
from qq.plugins.settings import SettingsBasedPlugin
from qq.plugins.settings import SettingsPlugin


class CeleryPlugin(SettingsBasedPlugin):
    DEFAULT_KEY = "celery"

    def __init__(
        self,
        celeryapp: Celery,
        settings_key: str = SettingsPlugin.DEFAULT_KEY,
    ):
        super().__init__(settings_key)
        self.celeryapp = celeryapp

    def start(self, application: Application):
        settings = self.get_my_settings(application)
        self.celeryapp.conf.update(settings)



from pyramid.config import Configurator

from qq.application import Application
from qq.plugins.settings import SettingsPlugin


class PyramidApplication(Application):
    PYRAMID_SETTINGS_KEY = "pyramid"
    _SETTINGS_KEY = SettingsPlugin.DEFAULT_KEY

    def make_wsgi_app(self, *args, **kwargs):
        """
        Configure application for web server and return pyramid's uwsgi
        application object.
        """
        pyramid = Configurator(*args, settings=self.globals[self._SETTINGS_KEY], **kwargs)
        pyramid.registry["application"] = self
        self._start_pyramid_plugins(pyramid)
        return pyramid.make_wsgi_app()

    def _start_pyramid_plugins(self, pyramid: Configurator):
        for plugin in self.plugins.values():
            method = getattr(plugin, "start_pyramid", lambda x: x)
            method(pyramid)



from typing import Any

from qq.application import Application
from qq.context import Context


class Plugin:
    def init(self, key: str):
        """
        Initialize the plguin during creating the plugins.
        key - key which is used in the Application.plugins dict for this plugin.
        """
        self.key = key

    def start(self, application: Application) -> Any:
        """
        This method will be called at the start of the Application. It will be
        called only once and the result will be set in the Application.globals.
        """

    def enter(self, context: Context) -> Any:
        """
        This method will be called when the Application will be used as context
        manager, but only when the plugin will be called. This is the enter phase.
        Result will be set in the Context dict with the self.key as the key in
        that dict.
        """

    def exit(self, context: Context, exc_type, exc_value, traceback):
        """
        This method will be called when the Application will be used as context
        manager. This is the exit phase.
        """



from qq.plugin import Plugin


class TestPlugin:
    """
    Plugin does not do anything by itself. This is only a sanity check.
    """

    def test_plugin(self):
        plugin = Plugin()
        plugin.start('configurator')
        plugin.enter('application')
        plugin.exit('application', 'exc_type', 'exc_value', 'traceback')



from abc import ABC
from abc import abstractmethod
from dataclasses import is_dataclass
from enum import Enum
from typing import List

from qq.plugins.jsonhack.encoders import DataclassEncoder
from qq.plugins.jsonhack.encoders import DateEncoder
from qq.plugins.jsonhack.encoders import DatetimeEncoder
from qq.plugins.jsonhack.encoders import DecimalEncoder
from qq.plugins.jsonhack.encoders import EnumEncoder
from qq.plugins.jsonhack.encoders import UUIDEncoder


class EncoderCreator(ABC):
    def __init__(self, customtype):
        self.customtype = customtype

    @abstractmethod
    def is_suitable(self):
        pass  # pragma: no cover

    @abstractmethod
    def create(self):
        pass  # pragma: no cover


class DataclassEncoderCreator(EncoderCreator):
    def is_suitable(self):
        return is_dataclass(self.customtype)

    def create(self):
        class ObjEncoder(DataclassEncoder):
            TYPE = self.customtype

        return ObjEncoder()


class CustomTypeEncoderCreator(EncoderCreator):
    base_encoders = (
        DatetimeEncoder,
        UUIDEncoder,
        DateEncoder,
        DecimalEncoder,
    )

    def __init__(self, customtype):
        super().__init__(customtype)
        self.base_encoder = None

    def is_suitable(self):
        for encoder in self.base_encoders:
            if issubclass(self.customtype, encoder().TYPE):
                self.base_encoder = encoder
                return True
        return False

    def create(self):
        class ObjEncoder(self.base_encoder):
            TYPE = self.customtype
            TYPENAME = self.customtype.__name__

        return ObjEncoder()


class EnumEncoderCreator(EncoderCreator):
    def is_suitable(self):
        return issubclass(self.customtype, Enum)

    def create(self):
        class ObjEncoder(EnumEncoder):
            TYPE = self.customtype
            TYPENAME = self.customtype.__name__

        return ObjEncoder()


def encoder_for(types: List[object]):
    creators = (
        DataclassEncoderCreator,
        CustomTypeEncoderCreator,
        EnumEncoderCreator,
    )
    for customtype in types:
        for creator in creators:
            creator = creator(customtype)
            if creator.is_suitable():
                yield creator.create()



from logging import getLogger

from tornado.ioloop import IOLoop

from qq.application import Application

log = getLogger(__name__)


class TornadoApplication(Application):
    def __init__(self):
        super().__init__()
        self.io_loop = IOLoop.instance()

    def start_tornado_loop(self):
        self._http_server.listen(self.settings["tornado_port"])
        try:
            log.info("Starting application...")
            self.io_loop.start()
        except KeyboardInterrupt:
            self.io_loop.stop()



import json
from dataclasses import dataclass

from qq.plugins.jsonhack.models import get_encoders
from qq.plugins.jsonhack.models import object_hook


@dataclass(init=False)
class EncoderStub:
    obj: object
    param_name: str
    default: object

    def __init__(self, obj: object, param_name: str):
        self.obj = obj
        self.param_name = param_name
        self.default = getattr(obj, param_name)

    def _get_function(self):
        def encoder(selfsecond, o):
            for encoder in get_encoders():
                if encoder.is_encodable(o):
                    return encoder.encode(o)
            return self.default(selfsecond, o)

        return encoder

    def stub(self):
        setattr(self.obj, self.param_name, self._get_function())

    def unstub(self):
        setattr(self.obj, self.param_name, self.default)


@dataclass(init=False)
class DecoderStub(EncoderStub):
    def _get_function(self):
        def decoder(selfsecond, *args, **kwargs):
            selfsecond.object_hook = object_hook
            selfsecond.scan_once = json.scanner.make_scanner(selfsecond)
            return self.default(selfsecond, *args, **kwargs)

        return decoder



from unittest.mock import MagicMock

from pytest import fixture
from pytest import raises

from qq.plugins.sqlalchemy.exceptions import SettingMissing
from qq.plugins.sqlalchemy.plugin import SqlAlchemyPlugin

PREFIX = "qq.plugins.sqlalchemy.plugin"


class TestSqlAlchemyPlugin:
    @fixture
    def mapp(self):
        return MagicMock()

    @fixture
    def mcreate_engine(self, mocker):
        return mocker.patch(f"{PREFIX}.create_engine")

    @fixture
    def mmake_url(self, mocker):
        return mocker.patch(f"{PREFIX}.make_url")

    @fixture
    def msessionmaker(self, mocker):
        return mocker.patch(f"{PREFIX}.sessionmaker")

    @fixture
    def plugin(self):
        plugin = SqlAlchemyPlugin()
        plugin.init("dbname")
        return plugin

    @fixture
    def mget_my_settings(self, mocker, plugin):
        mock = mocker.patch.object(plugin, "get_my_settings")
        mock.return_value = {
            "url": "sqlite:///tmp.first.db",
            "options": {"optionkey": "option value"},
        }
        return mock

    @fixture
    def metadata(self):
        return MagicMock()

    def test_start(
        self, plugin, mapp, msessionmaker, mcreate_engine, mget_my_settings
    ):
        """
        .start should create proper sqlalchemy engine.
        """
        plugin.start(mapp)

        mcreate_engine.assert_called_once_with(
            "sqlite:///tmp.first.db", optionkey="option value"
        )

        assert plugin.sessionmaker == msessionmaker.return_value
        msessionmaker.assert_called_once_with(
            autoflush=False, autocommit=False, bind=mcreate_engine.return_value
        )

    def test_enter(self, plugin):
        """
        .enter should create new database session and assign it to the context
        """
        plugin.sessionmaker = MagicMock()
        mcontext = MagicMock()
        plugin.engine = MagicMock()

        assert plugin.enter(mcontext) == plugin.sessionmaker.return_value

        plugin.sessionmaker.assert_called_once_with()

    def test_exit(self, plugin):
        """
        .exit should close the database session.
        """
        plugin.session = MagicMock()
        plugin._settings = {
            "url": "sqlite:///tmp.first.db",
            "options": {"optionkey": "option value"},
        }

        plugin.exit(None, None, None, None)

        assert not plugin.session.rollback.called
        plugin.session.close.assert_called_once_with()

    def test_exit_with_traceback(self, plugin):
        """
        .exit should rollback and close the database session when exception
        occured.
        """
        plugin.session = MagicMock()

        plugin.exit(None, True, None, None)

        plugin.session.rollback.assert_called_once_with()
        plugin.session.close.assert_called_once_with()

    def test_dbname(self, plugin, mapp, msessionmaker, mcreate_engine, mget_my_settings):
        """
        .dbname should return name of the database made from the db url
        """
        plugin.start(mapp)

        assert plugin.dbname == "tmp.first.db"

    def test_validate_settings(self, plugin, mapp, mget_my_settings):
        """
        Starting plugin should raise an error when settings are not properly
        configured (missing url)
        """
        del mget_my_settings.return_value["url"]
        with raises(SettingMissing):
            plugin.start(mapp)

    def test_recreate(self, plugin, metadata, mcreate_engine, mapp, mget_my_settings):
        """
        .recreate should drop all and create all tables using provided metadata.
        """

        plugin.start(mapp)
        plugin.recreate(metadata)

        metadata.drop_all.assert_called_once_with(mcreate_engine.return_value)
        metadata.create_all.assert_called_once_with(mcreate_engine.return_value)



from qq.plugin import Plugin
from qq.plugins.pyramid.application import PyramidApplication
from qq.plugins.settings import SettingsBasedPlugin


class BasePyramidPlugin(SettingsBasedPlugin):
    def start(self, application: PyramidApplication):
        self.settings = self.get_my_settings(application)


class AuthPlugin(BasePyramidPlugin):
    """
    Add authorization to the pyramid app.
    """

    def __init__(self, authn_policy_cls, authz_policy_cls, root_factory=None):
        self.authn_policy_cls = authn_policy_cls
        self.authz_policy_cls = authz_policy_cls
        self.root_factory = root_factory

    def start_pyramid(self, pyramid):
        authn_policy = self.authn_policy_cls(self.settings["secret"])
        authz_policy = self.authz_policy_cls()
        pyramid.set_authentication_policy(authn_policy)
        pyramid.set_authorization_policy(authz_policy)
        if self.root_factory:
            pyramid.set_root_factory(self.root_factory)


class CsrfPlugin(BasePyramidPlugin):
    """
    Add csrf mechanism to the pyramid app.
    """

    def __init__(self, policy_cls):
        self.policy_cls = policy_cls

    def start_pyramid(self, pyramid):
        pyramid.set_csrf_storage_policy(self.policy_cls())
        pyramid.set_default_csrf_options(
            require_csrf=True,
            token=self.settings["csrf_token_key"],
            header=self.settings["csrf_header_key"],
        )


class RoutingPlugin(Plugin):
    """
    Add routing to the pyramid app.
    """

    def __init__(self, routing_cls):
        self.routing_cls = routing_cls

    def start_pyramid(self, pyramid):
        self.routing = self.routing_cls(pyramid)
        self.routing.make()


class SessionPlugin(BasePyramidPlugin):
    """
    Add session mechanism to the pyramid app.
    """

    def __init__(self, session_factory_cls):
        self.session_factory_cls = session_factory_cls

    def start_pyramid(self, pyramid):
        secret = self.settings["session_secret"]
        session_factory = self.session_factory_cls(secret)
        pyramid.set_session_factory(session_factory)



from collections import OrderedDict
from contextvars import ContextVar

from qq.errors import AlreadyStartedError

KEY_PREFIX = "quackquack"


class Application:
    def __init__(self):
        self.is_started = False
        self.startpoint = None
        self.plugins = OrderedDict()
        self.extra = {}
        self.globals = {}
        self.context = ContextVar(self.context_var_key)
        self._token = None

    def start(self, startpoint: str = "default", **kwargs):
        """
        Start application. Return True if started. Return False if already
        started before.
        """
        if self.is_started:
            raise AlreadyStartedError()

        self.startpoint = startpoint
        self.extra = kwargs

        self.create_plugins()
        self._start_plugins()

        self.is_started = True

    def _start_plugins(self):
        for key, plugin in self.plugins.items():
            plugin.init(key)
            self.globals[key] = plugin.start(self)

    def create_plugins(self):
        pass

    @property
    def context_var_key(self):
        return f"{KEY_PREFIX}_{self.__class__.__name__}"



"""
This hack allows to encode uuid4, datetime and dataclases in json.
"""
from logging import getLogger
from typing import List

from pyramid.config import Configurator as PyramidConfigurator
from pyramid.renderers import JSON

from qq.application import Application
from qq.finder import ObjectFinder
from qq.plugin import Plugin
from qq.plugins.jsonhack.models import get_encoders
from qq.plugins.jsonhack.models import init_encoders
from qq.plugins.jsonhack.stubs import EncoderStub

logger = getLogger(__name__)


class JsonPlugin(Plugin):
    def __init__(
        self,
        stubs: List[EncoderStub],
        finders: List[ObjectFinder] = None,
    ):
        super().__init__()
        self.stubs = stubs
        self.finders = finders

    def start(self, application: Application):
        init_encoders(self.finders)
        for stub in self.stubs:
            stub.stub()

    def start_pyramid(self, pyramid: PyramidConfigurator):
        rendered = JSON()
        for encoder in get_encoders():
            rendered.add_adapter(
                encoder.TYPE, create_pyramid_json_adapter(encoder)
            )
        pyramid.add_renderer("json", rendered)


def create_pyramid_json_adapter(encoder):
    def encode(value, request):
        if encoder.is_encodable(value):
            return encoder.encode(value)

    return encode



from unittest.mock import MagicMock
from unittest.mock import patch

from pytest import fixture

from qq.plugins.pyramid.routing import Routing


class ExampleView:
    renderer = 'myrenderer'
    path_info = None


class TestRouting:
    @fixture
    def mconfig(self):
        return MagicMock()

    @fixture
    def routing(self, mconfig):
        return Routing(mconfig)

    @fixture
    def madd(self, routing):
        patcher = patch.object(routing, 'add')
        with patcher as mock:
            yield mock

    @fixture
    def madd_view(self, routing):
        patcher = patch.object(routing, 'add_view')
        with patcher as mock:
            yield mock

    @fixture
    def mread_from_file(self, routing):
        patcher = patch.object(routing, 'read_from_file')
        with patcher as mock:
            yield mock

    def test_add(self, routing, mconfig, madd_view):
        routing.add('view', 'route', 'url', 'arg', kw='arg')

        mconfig.add_route('route', 'url', 'arg', kw='arg')
        madd_view.assert_called_once_with('view', route_name='route')

    def test_add_view(self, routing, mconfig):
        mconfig.maybe_dotted.return_value = ExampleView
        routing.add_view(
            'impaf.tests.test_routing.ExampleView',
            route_name='something',
        )

        mconfig.add_view(
            'impaf.tests.test_routing.ExampleView',
            route_name='something',
            renderer='myrenderer',
        )



from redis import Redis

from qq.context import Context
from qq.plugins.settings import SettingsBasedPlugin

SETTINGS_HOST_KEY = "host"
SETTINGS_PORT_KEY = "port"
SETTINGS_DB_KEY = "db"


class RedisPlugin(SettingsBasedPlugin):
    def enter(self, context: Context):
        settings = self.get_my_settings(context)
        params = dict(
            host=settings[SETTINGS_HOST_KEY],
            port=settings[SETTINGS_PORT_KEY],
            db=settings[SETTINGS_DB_KEY],
        )
        return Redis(**params)



from contextvars import ContextVar
from unittest.mock import MagicMock
from unittest.mock import sentinel

from pytest import fixture

from qq.context import Context
from qq.injector import ContextManagerInjector
from qq.injector import InitializeInjectors
from qq.injector import Injector


class Fixtures:
    @fixture
    def mplugin(self):
        mock = MagicMock()
        mock.enter.return_value = sentinel.name
        return mock

    @fixture
    def mapp(self, mplugin):
        mock = MagicMock()
        mock.is_started = True
        mock.context = ContextVar("testing")
        mock.plugins = {
            "name": mplugin,
        }
        return mock

    @fixture
    def data(self):
        return {
            "name": 0,
        }

    @fixture
    def context(self, mapp):
        return Context(mapp)


class TestReturnInjector(Fixtures):
    @fixture
    def exampleinjector(self, data):
        @Injector
        def fun(context, key):
            data[key] = 1
            return context[key]

        return fun

    @fixture
    def examplefun(self, exampleinjector, mapp, data):
        def fun(obj=exampleinjector("name")):
            assert data["name"] == 1
            return obj

        return InitializeInjectors(mapp)(fun)

    def test_yielding(self, data, examplefun):
        assert data["name"] == 0
        assert examplefun() == sentinel.name
        assert data["name"] == 1


class TestYieldInjector(Fixtures):
    @fixture
    def exampleinjector(self, data):
        @Injector
        def fun(context, key):
            data[key] = 1
            yield context[key]
            data[key] = 2

        return fun

    @fixture
    def examplefun(self, exampleinjector, mapp, data):
        def fun(obj=exampleinjector("name")):
            assert data["name"] == 1
            return obj

        return InitializeInjectors(mapp)(fun)

    def test_yielding(self, data, examplefun):
        assert data["name"] == 0
        assert examplefun() == sentinel.name
        assert data["name"] == 2


class TestContextManagerInjector(Fixtures):
    @fixture
    def exampleinjector(self, data):
        class cls(ContextManagerInjector):
            def __enter__(self, context, key):
                data[key] = 1
                return context[key]

            def __exit__(self, exc_type, exc_value, traceback, context, key):
                data[key] = 2

        return cls

    @fixture
    def examplefun(self, exampleinjector, mapp, data):
        def fun(obj=exampleinjector("name")):
            assert data["name"] == 1
            return obj

        return InitializeInjectors(mapp)(fun)

    def test_context_manager(self, data, examplefun):
        assert data["name"] == 0
        assert examplefun() == sentinel.name
        assert data["name"] == 2



from contextlib import suppress
from functools import wraps
from inspect import BoundArguments
from inspect import signature
from sys import exc_info
from types import GeneratorType
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import List
from typing import Tuple

from qq.application import Application
from qq.context import Context


class InitializeInjectors:
    def __init__(self, application: Application):
        self.application = application

    def __call__(self, method: Callable) -> Callable:
        """
        If function has injectors in the defaults of arguments, then start them.
        """

        @wraps(method)
        def wrapper(*args, **kwargs):
            injector_list = list(
                get_injectors_from_function(method, args, kwargs)
            )
            for name, injector in injector_list:
                kwargs[name] = injector.start(wrapper._qq_application)

            try:
                return method(*args, **kwargs)
            finally:
                for name, injector in reversed(injector_list):
                    injector.end()

        def set_qq_app(application: Application):
            wrapper._qq_application = application

        wrapper._set_qq_app = set_qq_app
        wrapper._set_qq_app(self.application)
        return wrapper


class Injector:
    def __init__(self, fun):
        self.fun = fun

    def __call__(self, *args, **kwargs):
        return self.__class__(self.fun).init(*args, **kwargs)

    def init(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        return self

    def start(self, application: Application):
        self.context = Context(application)
        self.entered = self.context.__enter__()
        self.result = self.fun(self.entered, *self.args, **self.kwargs)
        if _is_generator(self.result):
            return self.result.__next__()
        else:
            return self.result

    def end(self):
        if _is_generator(self.result):
            with suppress(StopIteration):
                self.result.__next__()
        self.context.__exit__(*exc_info())


class ContextManagerInjector(Injector):
    def __init__(self, *args, **kwargs):
        super().__init__(None)
        self.init(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        return self  # pragma: no cover

    def start(self, application: Application):
        self.context = Context(application)
        self.entered = self.context.__enter__()
        return self.__enter__(self.entered, *self.args, **self.kwargs)

    def end(self):
        self.__exit__(
            *exc_info(),
            self.entered,
            *self.args,
            **self.kwargs,
        )
        self.context.__exit__(*exc_info())


@Injector
def SimpleInjector(context: Context, key: str):
    return context[key]


def is_injector_ready(parameter: Any, name: str, bound_args: BoundArguments):
    """
    Is parameter an injector and the value is not provided in the call.
    """
    is_injector = isinstance(parameter._default, Injector)
    return is_injector and name not in bound_args.arguments


def get_injectors_from_function(
    method: Callable, args: List, kwargs: Dict
) -> Iterable[Tuple[str, Callable]]:
    """
    Generate list of iterators that was not overwrited by function's call args.
    """
    sig = signature(method)
    bound_args = sig.bind_partial(*args, **kwargs)

    for name, parameter in sig._parameters.items():
        if is_injector_ready(parameter, name, bound_args):
            yield name, parameter._default


def _is_generator(obj):
    return isinstance(obj, GeneratorType)



from qq.plugins.jsonhack.encoders import Encoder


class MyInt(int):
    ENCODED_TYPENAME = "something"


class SampleEncoder(Encoder):
    TYPE = MyInt

    def _encode(self, value):
        pass  # pragma: no cover

    def _decode(self, value):
        pass  # pragma: no cover)


class TestEncoder:
    def test_typename(self):
        """
        .TYPENAME should return ENCODED_TYPENAME from .TYPE
        """
        SampleEncoder.TYPE = MyInt

        assert SampleEncoder().TYPENAME == "something"

    def test_typename2(self):
        """
        .TYPENAME should return __name__ from TYPE class
        """
        SampleEncoder.TYPE = int

        assert SampleEncoder().TYPENAME == "int"



from unittest.mock import MagicMock

from pytest import fixture


class ViewFixture:
    """
    In able to use this mixin, your test class needs to have fixture named "view".
    @fixture
    def view(self, mroot_factory, mrequest):
        return View(mroot_factory, mrequest)
    """

    _view = None

    @fixture
    def app(self):
        return MagicMock()

    @fixture
    def mroot_factory(self):
        return MagicMock()

    @fixture
    def mrequest(self, app):
        request = MagicMock()
        request.registry = dict(application=app)
        request._cache = {}
        request.GET = {}
        request.POST = {}
        request.matchdict = {}
        request.json_body = {}
        request.headers = {}
        return request

    @fixture
    def matchdict(self, mrequest):
        return mrequest.matchdict

    @fixture
    def view(self, mroot_factory, mrequest):
        return self._view(mroot_factory, mrequest)



from abc import ABC
from abc import abstractmethod
from dataclasses import is_dataclass
from importlib import import_module
from inspect import getmodule
from logging import getLogger
from pkgutil import walk_packages
from typing import List

from qq.plugins.types import CustomBaseType

logger = getLogger(__name__)


def is_defined_in(element, module):
    return getmodule(element).__name__ == module.__name__


class ObjectFinder(ABC):
    """
    This class is responsible for importing all modules in a project and search
    for specyfic objects.
    """

    process_cache = {}

    def __init__(
        self,
        parents: List[str],
        ignore_list: List[str] = None,
        cache_key: str = None,
    ):
        self.parents = parents
        self._cache_key = cache_key
        self.ignore_list = ignore_list or []

    @property
    def cache_key(self):
        return self._cache_key or self.__class__.__name__

    def find(self, force: bool = False):
        if force or self.cache_key not in self.process_cache:
            self.process_cache[self.cache_key] = self._find()
        return self.process_cache[self.cache_key]

    def _find(self):
        elements = []
        for parent in self.parents:
            info = f"{self.cache_key}: Searching for all objects in {parent}"
            logger.info(info)
            for package in self._get_all_packages(parent):
                elements += list(self._find_in_package(package))
        return elements

    def _get_all_packages(self, parent: str):
        try:
            parentpkg = import_module(parent)
        except Exception:
            logger.error(f"Can not import parent module: {parent}")
            return

        prefix = f"{parentpkg.__name__}."
        for module in walk_packages(parentpkg.__path__, prefix):
            if module.name in self.ignore_list:
                continue

            try:
                yield import_module(module.name)
            except Exception:
                logger.warning(f"Can not import module: {module.name}")
                continue

    def _find_in_package(self, package):
        intro = False
        for elementname in dir(package):
            element = getattr(package, elementname)

            if self.is_collectable(element) and is_defined_in(element, package):
                if not intro:
                    logger.debug(f"Module found: {package.__name__}")
                    intro = True
                logger.debug(f"\tObject found: {elementname}")
                yield element

    @abstractmethod
    def is_collectable(self, element: object):
        pass  # pragma: no cover


class DataclassFinder(ObjectFinder):
    def is_collectable(self, element: object):
        return is_dataclass(element)


class CustomBaseTypeFinder(ObjectFinder):
    def is_collectable(self, element: object):
        return issubclass(element, CustomBaseType)



from unittest.mock import MagicMock
from unittest.mock import sentinel

from pytest import fixture
from pytest import raises

from qq.application import AlreadyStartedError
from qq.application import Application
from qq.plugin import Plugin


class ExamplePlugin(Plugin):
    def enter(self, context):
        context["example"] = sentinel.example


class ExampleApplication(Application):
    def create_plugins(self):
        super().create_plugins()
        self.plugins["plugin1"] = MagicMock()
        self.plugins["plugin2"] = MagicMock()
        self.plugins["plugin3"] = ExamplePlugin()


class TestApplication:
    @fixture
    def app(self):
        return ExampleApplication()

    def test_start(self, app):
        """
        .start should append plugins and init them. Also proper flags should be
        set.
        """
        app.start("wsgi", wsgi=1)

        plugin1 = app.plugins["plugin1"]
        plugin2 = app.plugins["plugin2"]

        assert app.extra == {"wsgi": 1}
        assert app.globals == {
            "plugin1": plugin1.start.return_value,
            "plugin2": plugin2.start.return_value,
            "plugin3": None,
        }
        assert app.startpoint == "wsgi"
        assert app.is_started

        plugin1.start.assert_called_once_with(app)
        plugin2.start.assert_called_once_with(app)

    def test_start_when_started(self, app):
        """
        .start should do nothing if app already started
        """
        app.is_started = True

        with raises(AlreadyStartedError):
            assert app.start("wsgi", wsgi=1)

        assert app.globals == {}
        assert app.startpoint is None
        assert app.plugins == {}



class Routing:

    values_to_set = [
        'permission',
        'attr',
        'renderer',
        'http_cache',
        'wrapper',
        'decorator',
        'mapper',
        'context',
        'request_type',
        'request_method',
        'request_param',
        'match_param',
        'containment',
        'xhr',
        'accept',
        'header',
        'path_info',
        'check_csrf',
        'physical_path',
        'effective_principals',
        'custom_predicates',
        'predicates',
        'require_csrf',
    ]

    def __init__(self, pyramid):
        self.pyramid = pyramid

    def add(self, view, route, url, *args, **kwargs):
        """
        Add routing for view.
        - view: view class or dotted url to it
        - route: name for the route
        - url - url pattern
        """
        self.pyramid.add_route(route, url, *args, **kwargs)

        self.add_view(view, route_name=route)

    def add_view(self, view, **kwargs):
        """
        Add view/view handler.
        - view: view class or dotted url to it
        """
        view = self.pyramid.maybe_dotted(view)

        for name in self.values_to_set:
            self.set_view_config(kwargs, view, name)

        self.pyramid.add_view(view, **kwargs)

    def set_view_config(self, kwargs, view, name):
        value = getattr(view, name, None)
        if value:
            kwargs[name] = value



from qq import Application
from qq.errors import ApplicationNotStartedError


class Context:
    def __init__(self, application: Application):
        self.application = application
        self.values = {}
        self.token = None

    @property
    def globals(self):
        return self.application.globals

    @property
    def extra(self):
        return self.application.extra

    def __enter__(self):
        if not self.application.is_started:
            raise ApplicationNotStartedError(
                "Application is not started! Use Application.start(startpoint)"
            )
        try:
            return self.application.context.get()
        except LookupError:
            # I'm the first context, so I can return myself.
            self.token = self.application.context.set(self)
            return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.token:
            for key, plugin in reversed(self.application.plugins.items()):
                if key in self.values:
                    plugin.exit(self, exc_type, exc_value, traceback)
            self.application.context.reset(self.token)
            self.token = None

    def __getitem__(self, key: str):
        if key not in self.values:
            self.values[key] = self.application.plugins[key].enter(self)
        return self.values[key]



from logging.config import dictConfig

from qq.application import Application
from qq.plugins.settings import SettingsBasedPlugin


class LoggingPlugin(SettingsBasedPlugin):
    """
    Add logging configuration. Needs 'logging' value in settings.
    https://docs.python.org/3.6/library/logging.config.html#logging.config.dictConfig
    """

    DEFAULT_KEY = "logging"

    def start(self, application: Application):
        dictConfig(self.get_my_settings(application))



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['qq',
 'qq.plugins',
 'qq.plugins.celery',
 'qq.plugins.jsonhack',
 'qq.plugins.jsonhack.tests',
 'qq.plugins.pyramid',
 'qq.plugins.pyramid.tests',
 'qq.plugins.sqlalchemy',
 'qq.plugins.sqlalchemy.tests',
 'qq.plugins.tests',
 'qq.plugins.tornado',
 'qq.tests']

package_data = \
{'': ['*']}

install_requires = \
['marshmallow_dataclass>=8.0.0,<9.0.0']

extras_require = \
{'alembic': ['alembic>=1.0.9,<2.0.0'],
 'celery': ['celery>=5.0.0,<6.0.0'],
 'developer': ['alembic>=1.0.9,<2.0.0',
               'sqlalchemy>=1.3.3,<2.0.0',
               'pyramid>=1.10.4,<2.0.0',
               'redis>=3.2.1,<4.0.0',
               'celery>=5.0.0,<6.0.0'],
 'pyramid': ['pyramid>=1.10.4,<2.0.0'],
 'redis': ['redis>=3.2.1,<4.0.0'],
 'sqlalchemy': ['sqlalchemy>=1.3.3,<2.0.0']}

setup_kwargs = {
    'name': 'quackquack',
    'version': '1.0.2',
    'description': 'Quack Quack: A simple application framework',
    'long_description': '# Quack Quack\n\nIf it quacks like a quack, then it\'s a Quack Quack.\nVersion: 1.0.0\n\n# Table of Contents\n1. [Overview](#overview)\n2. [Quick Using Guide](#quick-using-guide)\n3. [Installation](#installation)\n4. [Tutorial](docs/tutorial.md)\n    * [Configuration](docs/tutorial.md#configuration)\n    * [Starting](docs/tutorial.md#starting)\n    * [Using Context](docs/tutorial.md#using-context)\n    * [Using Injectors](docs/tutorial.md#using-injectors)\n    * [Creating Plugins](docs/tutorial.md#creating-plugins)\n5. [Plugins](docs/plugins.md)\n    * [Settings](docs/plugins.md#settings)\n    * [Logging](docs/plugins.md#logging)\n    * [Redis](docs/plugins.md#redis)\n    * [Pyramid Plugin](docs/pyramid.md)\n    * [Sqlalchemy Plugin](docs/sqlalchemy.md)\n    * [Json Plugin](docs/json.md)\n6. [Phases](docs/phases.md)\n    * [About Phases](docs/phases.md#about-phases)\n    * [Phase 0](docs/phases.md#phase-0)\n    * [Phase 1 - creating Configurator instance](docs/phases.md#phase-1---creating-configurator-instance)\n    * [Phase 2 - starting Configurator](docs/phases.md#phase-2---starting-configurator)\n    * [Extending Phases](#extending-phases)\n    * [Application Phase Start](docs/phases.md#application-phase-start)\n    * [Application Phase End](docs/phases.md#application-phase-end)\n7. More info\n    * [Changelog](docs/CHANGELOG.md)\n8. [Example](example/readme.md)\n\n\n# Overview\n\nThis project aims to resolve problem of configuring an application, which needs to\nhave initialization step (for example: for gathering settings or establishing\nconnections) and use Python style code (context managers and decorators) to get\nthose data.\n\nFor example, normally you would need to use two separate mechanism for settings\nin celery application and web application, because you should not use web\napplication startup process in the celery app. This package provide a solution\nfor this problem, by giving one simple and independent of other frameworks\nmechanism to implement everywhere.\n\n# Quick Using Guide\n\nTo use Quack Quack you need to create the application class (inherited from\n`qq.Application`) in which you need to add plugins. After configuring, you need to "start"\nthe application. After that you can use the configurator as context manager.\n\n```python\nfrom qq import Application, Context\nfrom qq.plugins import SettingsPlugin\n\nclass MyApplication(Application):\n    def create_plugins(self):\n        self.plugins["settings"] = SettingsPlugin(\'esett\')\n\napplication = MyApplication()\napplication.start(\'application\')\n\nwith Context(application) as ctx:\n    print(ctx["settings"])\n\n```\n\n`context.settings` in above example is variable made by the SettingsPlugin.\nIf you would like to know more, please go to the [Tutorial](docs/tutorial.md)\n\n# Installation\n\n```bash\npip install quackquack\n```\n',
    'author': 'Dominik Dlugajczyk',
    'author_email': 'msocek@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/socek/quackquack',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from unittest.mock import MagicMock

from pytest import fixture

from qq.plugins.jsonhack.plugin import JsonPlugin
from qq.plugins.jsonhack.plugin import create_pyramid_json_adapter


class TestJsonPlugin:
    @fixture
    def mstub(self):
        return MagicMock()

    @fixture
    def mencoder(self):
        return MagicMock()

    @fixture
    def mpyramid(self):
        return MagicMock()

    @fixture
    def mget_encoders(self, mocker, mencoder):
        mock = mocker.patch("qq.plugins.jsonhack.plugin.get_encoders")
        mock.return_value = [mencoder]
        return mock

    @fixture
    def mcreate_pyramid_json_adapter(self, mocker):
        return mocker.patch("qq.plugins.jsonhack.plugin.create_pyramid_json_adapter")

    @fixture
    def mjson(self, mocker):
        return mocker.patch("qq.plugins.jsonhack.plugin.JSON")

    @fixture
    def plugin(self, mstub):
        return JsonPlugin([mstub])

    def test_start(self, plugin, mstub):
        """
        .start should start all stubs.
        """
        plugin.start(None)
        mstub.stub.assert_called_once_with()

    def test_start_pyramid(self, plugin, mget_encoders, mcreate_pyramid_json_adapter, mpyramid, mjson, mencoder):
        """
        .start_pyramid should create adapter from every encovder
        """
        plugin.start_pyramid(mpyramid)
        mpyramid.add_renderer.assert_called_once_with("json", mjson.return_value)
        mjson.return_value.add_adapter.assert_called_once_with(mencoder.TYPE, mcreate_pyramid_json_adapter.return_value)
        mcreate_pyramid_json_adapter.assert_called_once_with(mencoder)


class TestCreatePyramidJsonEncoder:
    @fixture
    def mencoder(self):
        return MagicMock()

    def test_create_pyramid_json_adapter_when_encoded(self, mencoder):
        """
        pyramid encoder works differently then normal encoders: encoder should
        return value only if the value can be encoded
        """
        mencoder.is_encodable.return_value = True

        pencoder = create_pyramid_json_adapter(mencoder)
        assert pencoder("x", None) == mencoder.encode.return_value
        mencoder.encode.assert_called_once_with("x")

    def test_create_pyramid_json_adapter_when_not_encoded(self, mencoder):
        """
        pyramid encoder works differently then normal encoders: encoder should
        return value only if the value can be encoded
        """
        mencoder.is_encodable.return_value = False

        pencoder = create_pyramid_json_adapter(mencoder)
        assert pencoder("x", None) is None
        assert not mencoder.encode.called



from unittest.mock import MagicMock
from unittest.mock import sentinel

from pytest import fixture
from pytest import raises

from qq.application import Application
from qq.context import ApplicationNotStartedError
from qq.context import Context
from qq.plugin import Plugin


class ExamplePlugin(Plugin):
    def enter(self, context):
        return {"one": sentinel.example, "two": sentinel.example2}


class ExampleApplication(Application):
    def create_plugins(self):
        super().create_plugins()
        self.plugins["plugin1"] = MagicMock()
        self.plugins["plugin2"] = MagicMock()
        self.plugins["plugin3"] = ExamplePlugin()


class TestContext:
    @fixture
    def app(self):
        return ExampleApplication()

    def test_when_not_started(self, app):
        """
        Application should raise an error when it will be used without starting
        it in the first place.
        """
        with raises(ApplicationNotStartedError):
            with Context(app):
                pass

    def test_context_names(self, app):
        """
        context.names() should return list of all vars in the context.
        """
        app.start()

        with Context(app) as context:
            assert (
                context["plugin1"] == app.plugins["plugin1"].enter.return_value
            )
            assert (
                context["plugin2"] == app.plugins["plugin2"].enter.return_value
            )
            assert context["plugin3"] == {
                "one": sentinel.example,
                "two": sentinel.example2,
            }

    def test_many_times(self, app):
        """
        Using app as context manager should enter and exit plugins only
        once.
        """
        app.start()

        with Context(app) as context1:
            with Context(app) as context2:
                assert not app.plugins["plugin1"].enter.called
                assert not app.plugins["plugin2"].enter.called

                assert context1["plugin1"] == context2["plugin1"]

        app.plugins["plugin1"].exit.assert_called_once_with(
            context1, None, None, None
        )
        assert not app.plugins["plugin2"].exit.called

    def test_when_error_raised(self, app):
        """
        When using app as context manager and an exception is raised,
        all plugins should get that exception.
        """
        app.start()
        error = RuntimeError()

        with raises(RuntimeError):
            with Context(app) as context:
                context["plugin1"]
                raise error

        app.plugins["plugin1"].start.assert_called_once_with(app)
        app.plugins["plugin2"].start.assert_called_once_with(app)

        call = app.plugins["plugin1"].exit.call_args_list[0][0]
        assert call[0] == context
        assert call[1] == RuntimeError
        assert call[2] == error
        assert call[3] is not None

        assert not app.plugins["plugin2"].exit.called

    def test_extra(self, app):
        app.start(kw=1)

        with Context(app) as ctx:
            assert ctx.extra == {"kw": 1}



from unittest.mock import MagicMock

from pytest import fixture

from qq.plugins.jsonhack.plugin import JsonPlugin


class TestJsonPlugin:
    @fixture
    def configurator(self):
        return MagicMock()

    @fixture
    def pyramid(self):
        return MagicMock()

    @fixture
    def stub(self):
        return MagicMock()

    @fixture
    def finder(self):
        return MagicMock()

    @fixture
    def minit_encoders(self, mocker, finder):
        return mocker.patch("qq.plugins.jsonhack.plugin.init_encoders")

    @fixture
    def mget_encoders(self, mocker):
        return mocker.patch("qq.plugins.jsonhack.plugin.get_encoders")

    @fixture
    def mrenderer(self, mocker):
        return mocker.patch("qq.plugins.jsonhack.plugin.JSON")

    @fixture
    def mcreate_pyramid_json_adapter(self, mocker):
        return mocker.patch("qq.plugins.jsonhack.plugin.create_pyramid_json_adapter")

    @fixture
    def plugin(self, stub, finder):
        return JsonPlugin([stub], finder)

    def test_start(self, plugin, configurator, minit_encoders, stub, finder):
        """
        .start should start all stubs
        """
        plugin.start(configurator)

        minit_encoders.assert_called_once_with(finder)
        stub.stub.assert_called_once_with()

    def test_start_pyramid(self, plugin, pyramid, mget_encoders, mrenderer, mcreate_pyramid_json_adapter):
        """
        .start_pyramid should append JSON renderer to pyramid's confiugration
        """
        encoder = MagicMock()
        encoder.is_encodable.return_value = True
        mget_encoders.return_value = [encoder]
        plugin.start_pyramid(pyramid)

        mrenderer.assert_called_once_with()
        mrenderer.return_value.add_adapter(encoder.TYPE, mcreate_pyramid_json_adapter.return_value)
        mcreate_pyramid_json_adapter.assert_called_once_with(encoder)



"""
This hack allows to encode uuid4 and datetime in json
"""
from inspect import isabstract
from logging import getLogger
from typing import List

from qq.plugins.jsonhack import encoders
from qq.plugins.jsonhack.encodergenerators import encoder_for
from qq.plugins.jsonhack.errors import UnknownObjectError

logger = getLogger(__name__)

ENCODERS_CACHE = set()


def get_encoders():
    return ENCODERS_CACHE


def is_encoder(obj: object) -> bool:
    return issubclass(obj, encoders.Encoder) and not isabstract(obj)


def init_encoders(finders: List = None):
    """
    Get all encoders from qq.plugins.jsonhack.encoders and from all dataclasses
    from the provided finder.
    """
    finders = finders or []
    for name in dir(encoders):
        if name.startswith("_"):
            continue
        element = getattr(encoders, name)
        try:
            if is_encoder(element):
                add_encoder(element())
        except TypeError:
            continue

    for finder in finders:
        # search for all dataclasses and generate encoder
        for encoder in encoder_for(finder.find()):
            add_encoder(encoder)


def object_hook(obj):
    if "_type" not in obj:
        return obj

    for encoder in get_encoders():
        if encoder.is_decodable(obj):
            return encoder.decode(obj)
    raise UnknownObjectError(obj)


def add_encoder(encoder):
    ENCODERS_CACHE.add(encoder)



from dataclasses import dataclass
from unittest.mock import MagicMock

from pytest import fixture

from qq.finder import CustomBaseTypeFinder
from qq.finder import DataclassFinder
from qq.finder import ObjectFinder
from qq.plugins.types import CustomBaseType


@dataclass
class SimpleDataclass:
    pass


class CustomStr(str, CustomBaseType):
    pass


class FakeFinder(ObjectFinder):
    def is_collectable(self, element: object):
        return type(element) is str


class TestObjectFinder:
    @fixture
    def finder(self):
        return FakeFinder(["something"])

    @fixture
    def mmodule(self):
        mmodule = MagicMock()
        mmodule.__name__ = "something"
        mmodule.__path__ = ["/somewhere/something"]
        return mmodule

    @fixture
    def melement(self):
        return MagicMock()

    @fixture
    def mwalk_packages(self, mocker, mmodule):
        mock = mocker.patch("qq.finder.walk_packages")
        mock.return_value = [mmodule]
        return mock

    @fixture
    def mimport_module(self, mocker, mmodule):
        mimport_module = mocker.patch("qq.finder.import_module")
        mimport_module.return_value = mmodule
        return mimport_module

    @fixture
    def mis_defined_in(self, mocker):
        return mocker.patch("qq.finder.is_defined_in")

    @fixture
    def mfind(self, mocker, finder):
        return mocker.patch.object(finder, "_find")

    @fixture
    def mfind_in_package(self, mocker, finder, melement):
        mock = mocker.patch.object(finder, "_find_in_package")
        mock.return_value = [melement]
        return mock

    def test_find_when_no_cache(self, finder, mfind):
        """
        .find should do full find if cache is empty
        """
        finder.process_cache = {}

        assert finder.find() == mfind.return_value
        mfind.assert_called_once_with()

    def test_find_when_find_forced(self, finder, mfind):
        """
        .find should do full find if cache is empty
        """
        finder.process_cache = {finder.cache_key: "something"}

        assert finder.find(True) == mfind.return_value
        mfind.assert_called_once_with()

    def test_find_when_cache(self, finder, mfind):
        """
        .find should do full find if cache is empty
        """
        finder.process_cache = {finder.cache_key: "something"}

        assert finder.find() == "something"
        assert not mfind.called

    def test_find_raw_when_ok(
        self,
        finder,
        mwalk_packages,
        mimport_module,
        mfind_in_package,
        mmodule,
        melement,
    ):
        """
        ._find should walk thru all packages and find all elements in those packages
        """
        assert finder._find() == [melement]

    def test_find_raw_when_not_ok(
        self,
        finder,
        mwalk_packages,
        mimport_module,
        mfind_in_package,
        mmodule,
        melement,
    ):
        """
        ._find should ignore package which can not be imported
        """
        mimport_module.side_effect = RuntimeError()
        assert finder._find() == []

    def test_find_raw_when_ignored(
        self,
        finder,
        mwalk_packages,
        mimport_module,
        mfind_in_package,
        mmodule,
        melement,
    ):
        """
        ._find should ignore package which are in ignored list
        """
        finder.ignore_list = [mmodule.name]
        assert finder._find() == []

    def test_find_in_package(self, finder, mis_defined_in):
        package = MagicMock()
        package.__name__ = "something"
        package.two = "something2"
        assert list(finder._find_in_package(package)) == [
            "something",
            "something2",
        ]


class TestDataclassFinder:
    def test_is_collectable(self):
        """
        DataclassFinder.is_collectable should return True only if element is an dataclass
        """
        finder = DataclassFinder(["something"])
        assert finder.is_collectable(SimpleDataclass) is True
        assert finder.is_collectable(str) is False


class TestCustomBaseTypeFinder:
    def test_is_collectable(self):
        """
        CustomBaseTypeFinder.is_collectable should return True only if element is an dataclass
        """
        finder = CustomBaseTypeFinder(["something"])
        assert finder.is_collectable(CustomStr) is True
        assert finder.is_collectable(str) is False
